PROCEDURE Newext (VAR Fnam: Filename; Ext: Fileext);
   EXTERNAL;

PROCEDURE Addext (VAR Fnam: Filename; Ext: Fileext);
   EXTERNAL;

PROCEDURE GetCommandline (VAR Commandline: Commandrec);
   EXTERNAL;

PROCEDURE Uexit (Errors: Boolean);
   EXTERNAL;

