/* dispatch numbers for 8086 instructions.  These numbers are placed in
 * the code_i component of each instruction bucket by init.c and are used
 * by ins.c for dispatching.
 */

#define i_aaa	1
#define i_aad	2
#define i_aam	3
#define i_aas	4
#define i_adc	5
#define i_add	6
#define i_and	7
#define i_call	8
#define i_calli	9
#define i_cbw	10
#define i_clc	11
#define i_cld	12
#define i_cli	13
#define i_cmc	14
#define i_cmp	15
#define i_cmpb	16
#define i_cmpw	17
#define i_cwd	18
#define i_daa	19
#define i_das	20
#define i_dec	21
#define i_div	22
#define i_esc	23
#define i_idiv	24
#define i_imul	25
#define i_in	26
#define i_inc	27
#define i_int	28
#define i_into	29
#define i_inw	30
#define i_iret	31
#define i_hlt	32
#define i_j	33
#define i_je	34
#define i_jl	35
#define i_jle	36
#define i_jb	37
#define i_jbe	38
#define i_jp	39
#define i_jo	40
#define i_js	41
#define i_jne	42
#define i_jnl	43
#define i_jnle	44
#define i_jnb	45
#define i_jnbe	46
#define i_jnp	47
#define i_jno	48
#define i_jns	49
#define i_jcxz	50
#define i_jmp	51
#define i_jmpi	52
#define i_lahf	53
#define i_lds	54
#define i_lea	55
#define i_les	56
#define i_lock	57
#define i_lodb	58
#define i_lodw	59
#define i_loop	60
#define i_loope	61
#define i_loopne	62
#define i_mov	63
#define i_movb	64
#define i_movw	65
#define i_mul	66
#define i_neg	67
#define i_not	68
#define i_or	69
#define i_out	70
#define i_outw	71
#define i_pop	72
#define i_popf	73
#define i_push	74
#define i_pushf	75
#define i_rcl	76
#define i_rcr	77
#define i_repnz	78
#define i_repz	79
#define i_ret	80
#define i_reti	81
#define i_rol	82
#define i_ror	83
#define i_sahf	84
#define i_sal	85
#define i_sar	86
#define i_sbb	87
#define i_scab	88
#define i_scaw	89
#define i_seg	90
#define i_shr	91
#define i_stc	92
#define i_std	93
#define i_sti	94
#define i_stob	95
#define i_stow	96
#define i_sub	97
#define i_test	98
#define i_wait	99
#define i_xchg	100
#define i_xlat	101
#define i_xor	102

#define i_beq	103
#define i_bne	104
#define i_ble	105
#define i_bgt	106
#define i_blt	107
#define i_bge	108
#define i_blos	109
#define i_bhi	110
#define i_blo	111
#define i_bhis	112
#define	i_br	113

#define i_addb	114
#define i_adcb	115
#define i_incb	116
#define i_subb	117
#define i_sbbb	118
#define i_decb	119
#define i_negb	120
#define i_mulb	121
#define i_imulb	122
#define i_divb	123
#define i_idivb	124
#define i_notb	125
#define i_salb	126
#define i_shrb	127
#define i_sarb	128
#define i_rolb	129
#define i_rorb	130
#define i_rclb	131
#define i_rcrb	132
#define i_andb	133
#define i_testb	134
#define i_orb	135
#define i_xorb	136

#define i_fabs		137
#define i_faddf		138
#define i_faddl		139
#define i_faddd		140
#define i_faddi		141
#define i_fadd		142
#define i_faddp		143
#define i_fchs		146
#define i_fclex		147
#define i_fcomf		148
#define i_fcoml		149
#define i_fcomd		150
#define i_fcomi		151
#define i_fcom		152
#define i_fcompf	153
#define i_fcompl	154
#define i_fcompd	155
#define i_fcompi	156
#define i_fcomp		157
#define i_fcompp	158
#define i_fdecstp	159
#define i_fdisi		160
#define i_fdivf		161
#define i_fdivl		162
#define i_fdivd		163
#define i_fdivi		164
#define i_fdiv		165
#define i_fdivp		166
#define i_fdivr		167
#define i_fdivrp	168
#define i_feni		169
#define i_ffree		170
#define i_fincstp	171
#define i_finit		172
#define i_fldcw		173
#define i_fldenv	174
#define i_fldf		175
#define i_fldl		176
#define i_fldd		177
#define i_fldi		178
#define i_fld		179
#define i_fldlg2	180
#define i_fldln2	181
#define i_fldl2e	182
#define i_fldl2t	183
#define i_fldpi		184
#define i_fldz		185
#define i_fld1		186
#define i_fnop		187
#define i_fmulf		188
#define i_fmull		189
#define i_fmuld		190
#define i_fmuli		191
#define i_fmul		192
#define i_fmulp		193
#define i_fpatan	196
#define i_fprem		197
#define i_fptan		198
#define i_frstor	199
#define i_frndint	200
#define i_fsave		201
#define i_fscale	202
#define i_fsqrt		203
#define i_fstf		204
#define i_fstl		205
#define i_fstd		206
#define i_fsti		207
#define i_fst		208
#define i_fstpf		209
#define i_fstpl		210
#define i_fstpd		211
#define i_fstpi		212
#define i_fstp		213
#define i_fstcw		214
#define i_fstenv	215
#define i_fstsw		216
#define i_fsubf		217
#define i_fsubl		218
#define i_fsubd		219
#define i_fsubi		220
#define i_fsub		221
#define i_fsubp		222
#define i_fsubr		223
#define i_fsubrp	224
#define i_ftst		225
#define i_fwait		226
#define i_fxam		227
#define i_fxch		228
#define i_fxtract	229
#define i_fyl2x		230
#define i_fyl2xp1	231
#define i_f2xm1		232
#define i_fdivrf	233
#define i_fdivrl	234
#define i_fdivrd	235
#define i_fdivri	236
#define i_fsubrf	237
#define i_fsubrl	238
#define i_fsubrd	239
#define i_fsubri	240

#define i_long	241
#define i_word	242
#define i_byte	243
#define i_text	244
#define i_data	245
#define i_bss	246
#define i_globl 247
#define i_comm	248
#define i_even	249
#define i_ascii 250
#define i_asciz 251
#define i_zerow 252
/* the following definition added by Bill O'Farrell June 28, 1982 */
#define i_para  253
