    char *pc0 = "macro processor - optional?";
    char *pc1 = "/usr/local/bin/upas";
    char *pc2 = "/usr/local/bin/lopt68";
    char *pc3 = "/usr/local/bin/ugen68";
    char *help = "/usr/sun/src/cmd/pc68/how_pc68";
    char *ld = "/usr/local/bin/ld68";
    char *dl = "/usr/local/bin/dl68";	    /* MACSBUG downloader */
    char *rev = "/usr/local/bin/rev68";    /* Pratt's .b 68k reformatter */
    char *as = "/usr/local/bin/as68";
    char *cp = "/bin/cp";
    char *mv = "/bin/mv";
